package iteriere;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.function.Consumer;
import java.util.Comparator;
import java.util.Iterator;
import java.util.NoSuchElementException;


public sealed interface LL<E> 
  extends Iterable<E>
	permits LL.Nil, LL.Cons{

	@Override
	default Iterator<E> iterator() {
		var xs = new LLSchleife<>(this);
		return xs;
	}

  static public final record Nil<E>() implements LL<E>{
    @Override public String toString(){return "[]";}
  }

  static public final record Cons<E>(E hd,LL<E> tl) implements LL<E>{
    @Override public String toString(){return show();}
  }

  default boolean isEmpty(){return this instanceof Nil;}

  default E head(){
    if (this instanceof Cons<E> c) return c.hd();
    throw new NoSuchElementException("head on empty list");
  }

  default LL<E> tail(){
    if (this instanceof Cons<E> c) return c.tl();
    throw new NoSuchElementException("tail on empty list");
  }

  static final LL nil = new Nil<>();
  static <E>LL<E> nil(){return nil;}

  static <E>LL<E> cons(E hd,LL<E> tl){return new Cons<>(hd,tl);}

  static <E>LL<E> of(E...es){
    LL<E> r = nil();
    for (int i=es.length-1;i>=0;i--) r = cons(es[i],r);
    return r; 
  }

  default String show(){
    StringBuffer result = new StringBuffer("[");
    boolean first = true;
    for (var it = this;!it.isEmpty();it=it.tail()){
      if (first) first = false;else result.append(", ");
      result.append(it.head());
    }
    result.append("]");
    return result.toString();
  }

  
  default LL<E>f(){
	  if (isEmpty()||tail().isEmpty())return nil();
	  var x1 =head();
//	  var x2 = tail().head();
	  var xs = tail().tail();
	  return cons(x1,xs.f());
  }
  
  default int lengthRec(){
    if (isEmpty()) return 0;
    return 1+tail().lengthRec();
  }
  default int length(){
	  int i = 0;
	  for (var x : this)i++;
	  return i;
  }
  default E lastRec(){
	if (isEmpty()) throw new NoSuchElementException("last on empty list");  
    if (tail().isEmpty()) return head();
    return tail().lastRec();
  }
  default E last(){
	  if (isEmpty()) throw new NoSuchElementException("last on empty list");  
	  E r = null;
	  for (var x: this) r = x;
	  return r;
  }

  default LL<E> appendIt(LL<E> that){
	  LL<E> rs = nil();
	  for (var x: this) rs = cons(x,rs);
	  for (var x: that) rs = cons(x,rs);
	  return rs.reverse();
  }
  default LL<E> append(LL<E> that){
    if (isEmpty()) return that;
    return cons(head(),tail().append(that)); 
  }

  default LL<E> drop(int i){
    if (i<=0 || isEmpty()) return this;
    return tail().drop(i-1);
  }
    
  default LL<E> take(int i){
    if (isEmpty()||i<=0) return nil(); 
    return cons(head(),tail().take(i-1));
  }

  default LL<E> sublist(int from, int length) {
    return drop(from).take(length);
  }

  default LL<E> reverse(){
    if (isEmpty()) return nil();
    return tail().reverse().append(of(head()));
  }

  default LL<E> intersperse(E e){
    if (isEmpty()||tail().isEmpty()) return this;
    return cons(head(),cons(e,tail().intersperse(e)));
  }

  default boolean isPrefixOf(LL<E> that){
    if (isEmpty()) return true;
    if (that.isEmpty()) return false;
    return head().equals(that.head())
    		&&tail().isPrefixOf(that.tail());
  }

  default boolean isSuffixOf(LL<E> that){
    return this.reverse().isPrefixOf(that.reverse());
  }

  default boolean isInfixOf(LL<E> that){
    return isPrefixOf(that)||
      (!that.isEmpty()&&isInfixOf(that.tail()));
  }

  default E getIt(int i){
    for (var x:this) {
    	if (i==0) return x;
    	i = i-1;
    }
    throw new IndexOutOfBoundsException();
  } 
  default E get2(int i){
    if (isEmpty()||i<0) throw new IndexOutOfBoundsException();
    if (i==0) return head();
    return tail().get2(i-1);
  }      
  default E get(int i){
    return drop(i).head();
  }



  default LL<E> rotate(){return tail().append(of(head()));}

  default LL<LL<E>> tails(){
    if (isEmpty()) return of(nil());
    return cons(this,tail().tails());
  }
  
  default void forEachRec(Consumer<? super E> con) {
    if (isEmpty()) return;
    con.accept(head());
    tail().forEachRec(con);
  }

  default boolean containsWithRec(Predicate< ? super E> p) {
    if (isEmpty()) return false;
    return p.test(head())||tail().containsWithRec(p);
  }
  default boolean containsWith(Predicate< ? super E> p) {
	  for (var x:this)
		  if (p.test(x))
			  return true;
	  return false;
  }
  default boolean contains(E el) {
    return containsWithRec(x->x.equals(el));
  }

  default LL<E> dropWhile(Predicate< ? super E> p){
    if (!isEmpty() && p.test(head()))
      return tail().dropWhile(p);
    return this;
  }

  default LL<E> takeWhile(Predicate< ? super E> p){
    if (!isEmpty() && p.test(head()))
      return cons(head(),tail().takeWhile(p));
    return nil();
  }

  default LL<E> filter(Predicate<? super E> p){
    if (isEmpty()) return nil();
    if (p.test(head())) return cons(head(), tail().filter(p));
    return tail().filter(p);
  }

  default <R> LL<R> map(Function<? super E, ? extends R> f){
    if (!isEmpty())
      return cons(f.apply(head()), tail().map(f));
    return nil();    
  }

  static public record Pair<A,B>(A fst,B snd){
    @Override public String toString(){return "("+fst()+", "+snd()+")";}
  }    

  default <B> LL<Pair<E,B>> zip(LL<B> that){
    if (this.isEmpty()||that.isEmpty()) return nil();
    return cons(new Pair<>(this.head(),that.head())
    		,this.tail().zip(that.tail()));
  }

  default Pair<LL<E>,LL<E>> span2(Predicate<? super E> p){
    if (isEmpty()) return new Pair<>(nil(),nil()); 
    if (!p.test(head())) return new Pair<>(nil(),this);
    var r = tail().span2(p);
    return new Pair<>(cons(head(),r.fst()),r.snd());
  }
  default Pair<LL<E>,LL<E>> span(Predicate<? super E> p){
    return new Pair<>(takeWhile(p),dropWhile(p));
  }

  default Pair<LL<E>,LL<E>> partition2(Predicate<? super E> p){
    if (isEmpty()) return new Pair<>(nil(),nil()); 
    var r = tail().partition2(p);
    return p.test(head())
      ?new Pair<>(cons(head(),r.fst()),r.snd())
      :new Pair<>(r.fst(),cons(head(),r.snd()))
    ;
  }
  default Pair<LL<E>,LL<E>> partition(Predicate<? super E> p){
    return new Pair<>(filter(p),filter(p.negate()));
  }
  default boolean isSorted(Comparator<? super E> cmp){
    return isEmpty()
     ||tail().isEmpty()
     ||cmp.compare(head(),tail().head())<=0 
     && tail().isSorted(cmp);
  }

  default LL<E> qsort(Comparator<? super E> cmp){
    if (isEmpty()) return nil();
    var p = tail().partition(x->cmp.compare(head(),x)>=0);
    return p.fst().qsort(cmp).append(cons(head(),p.snd().qsort(cmp)));
  }
}